// 4004 peripherals emulator and tester
// Copyright 2012 by Frank Buss, fb@frank-buss.de

#include <stdio.h>
#include "hardware.h"
#include "lcd.h"

// PIC timer
uint16_t g_tick = 0;
uint8_t g_second = 0;

// clock generator variables
uint8_t g_phase = 0;
uint8_t g_cycle = 0;
uint8_t g_cycleReset = 0;

// requested ROM address from the 4004
uint8_t g_address = 0;

// counter for sync signals
uint32_t g_syncCount = 0;

// nop
// nop
// nop
// nop
// jun 0
uint8_t g_romLoop[] = {
	0x00, 0x00, 0x00, 0x00, 0x40, 0x00 
};

// program http://www.frank-buss.de/4004/led-test.txt
uint8_t g_rom[] = {
	0xD0, 0xFD, 0x20, 0x00, 0x21, 0xF0, 0xB2, 0xD1,
	0xB3, 0xA3, 0xE1, 0xA2, 0x14, 0x1F, 0xD1, 0xB4,
	0xA3, 0xF1, 0x94, 0x1C, 0x19, 0xD0, 0xB2, 0x40,
	0x09, 0xA3, 0xF1, 0xF6, 0xB3, 0x40, 0x09, 0xD8,
	0xB4, 0xA3, 0xF1, 0x94, 0x1C, 0x2A, 0xD1, 0xB2,
	0x40, 0x09, 0xA3, 0xF1, 0xF5, 0xB3, 0x40, 0x09
};

// interrupt handler (called at 4 kHz)
static void interrupt isr(void)
{
	// clear interrupt flag
	TMR2IF = 0;

	// reset CPU, if button 1 is pressed
	setCpuReset(isButton1Pressed());

	// update second counter and LED
	g_tick++;
	if (g_tick == 4000) {
		g_tick = 0;
		g_second++;
		if (g_second == 100) g_second = 0;
	}

	// LED blinking
	setRedLed(g_tick < 1000);

	// 4004 clock generator
	switch (g_phase) {
		case 0:
			setPhi1(1);
			setPhi2(1);

			// reset or update cycle counter
			if (g_cycleReset) {
				g_cycle = 0;
				g_cycleReset = 0;
			} else {
				g_cycle++;
			}
			break;

		case 1:
			setPhi1(0);
			setPhi2(1);
			break;
	
		case 2:
			setPhi1(1);
			setPhi2(1);

			// the 4004 sets sync to low on rising edge of phi2, test on rising edge of phi1
			if (isCpuSync()) {
				// schedule reset cycle counter on start of cycle 0
				g_cycleReset = 1;
				g_syncCount++;
			}
			break;
	
		case 3:
			setPhi1(1);
			setPhi2(0);
			break;

		default:
			break;
	}

	// 4004 ROM emulation
	// the CPU sends 12 bits of address in the first three cycles (A1, A2, A3)
	// the ROM sends back 8 bits back in the next two cycles (M1, M2)

	// sample data lines on phase 2
	if (g_phase == 2) {
		switch (g_cycle) {
			// A1: lower 4 bit of address
			case 0:
				g_address = getCpuData();
				break;

			// A1: middle 4 bit of address
			case 1:
				g_address |= getCpuData() << 4;
				break;

			// A2: higher 4 bit of address (chip select code)
			case 2:
				break;
		}
	}

	// send data on phase 0
	if (g_phase == 0) {
		switch (g_cycle) {
			// M1, OPR (upper 4 bits)
			case 3:
				setCpuEnableDataOut(1);
				setCpuData(g_rom[g_address] >> 4);
				break;

			// M2, OPA (lower 4 bits)
			case 4:
				setCpuData(g_rom[g_address]);
				break;

			// disable output at next cycle
			case 5:
				setCpuEnableDataOut(0);
				break;
		}
	}

	// next clock phase
	g_phase++;
	if (g_phase == 4) g_phase = 0;
}

// main function
void main(void)
{
	char buf[17];

	// init hardware, set reset line of the 4004 CPU
	hardwareInit();
	lcdInit();
	setCpuReset(1);

	// show info
	drawTextLine(0, "4004 Emulator");
	drawTextLine(1, "by Frank Buss");
	__delay_ms(3000);
	lcdClear();

	// enable interrupts
	enableInterrupt();
	
	// display loop, the rest is in the interrupt function
	uint8_t oldSecond = 254;
	while (1) {
		if (oldSecond != g_second) {
			sprintf(buf, "address: 0x%02x", g_address);
			drawTextLine(0, buf);
			sprintf(buf, "sync: %lu", g_syncCount);
			drawTextLine(1, buf);
			oldSecond = g_second;
		}
	}
}
